#include <msp430.h>
#include <stdint.h>
#include <string.h>
#include <stdbool.h>

#define BOARD_CS_MCLK_FLLREF_RATIO_2MHZ                 (60)
#define BOARD_CS_MCLK_FLLREF_RATIO_4MHZ                 (121)
#define BOARD_CS_MCLK_FLLREF_RATIO_8MHZ                 (243)
#define BOARD_CS_MCLK_FLLREF_RATIO_16MHZ                (487)
#define BOARD_CS_MCLK_FLLREF_RATIO_24MHZ                (731)
#define BOARD_CS_SMCLK_2MHZ_CALCULATED_FREQUENCY_IN_HZ  ((BOARD_CS_MCLK_FLLREF_RATIO_2MHZ + 1) * 32768)
#define BOARD_CS_SMCLK_24MHZ_CALCULATED_FREQUENCY_IN_HZ ((BOARD_CS_MCLK_FLLREF_RATIO_24MHZ + 1) * 32768)
#define BOARD_CS_MCLK_2MHZ_CALCULATED_FREQUENCY_IN_KHZ  (BOARD_CS_SMCLK_2MHZ_CALCULATED_FREQUENCY_IN_HZ / 1000)
#define BOARD_CS_MCLK_24MHZ_CALCULATED_FREQUENCY_IN_KHZ (BOARD_CS_SMCLK_24MHZ_CALCULATED_FREQUENCY_IN_HZ / 1000)


#define FIN1_IO_PORT_DIR      P1DIR                          /**< Portzuordnung Richtungsregister */
#define FIN1_IO_PORT_SEL0     P1SEL0                         /**< Portzuordnung Auswahlregister 1*/
#define FIN1_IO_PORT_SEL1     P1SEL1                         /**< Portzuordnung Auswahlregister 2*/
#define FIN1_IO               BIT0                           /**< Pin Zuordnung  */
#define FIN1_IO_SET_SET_AS_IN FIN1_IO_PORT_DIR &= ~(FIN1_IO) /**< Pin als Eingang setzen */

#define FIN1_COMPARATOR_CTL_REG      CP0CTL0    /**< Comparator Control Register 0 */
#define FIN1_COMPARATOR_CTL1_REG     CP0CTL1    /**< Comparator Control Register 1 */
#define FIN1_COMPARATOR_DAC_CTL_REG  CP0DACCTL  /**< 6-bit Comparator built-in DAC Control Register */
#define FIN1_COMPARATOR_DAC_DATA_REG CP0DACDATA /**< 6-bit Comparator built-in DAC Data Register */


typedef enum {
    CLOCK_DIV_1 = 1,
    CLOCK_DIV_2 = 2,
    CLOCK_DIV_4 = 4,
    CLOCK_DIV_8 = 8
} CLOCK_DIV;

#define ACLK  32.768E3F
#define SMCLK 24.00E6F

#define FOUT_TIMER_CCR0_REG TB1CCR0
#define FOUT_TIMER_CCR1_REG TB1CCR1
#define FOUT_TIMER_CTL_REG  TB1CTL
#define FOUT_TIMER_CCTL_REG TB1CCTL1

#define FOUT_TIMER_PWM_MODE_TOGGLE_SET OUTMOD_6
#define FOUT_TIMER_PWM_MODE_OUT        OUTMOD_0

#define FOUT_IO_PORT_DIR       P2DIR                          /**< Portzuordnung Richtungsregister */
#define FOUT_IO_PORT_OUT       P2OUT                          /**< Portzuordnung Ausgangsregister */
#define FOUT_IO_PORT_SEL0      P2SEL0                         /**< Portzuordnung Ausgangsregister */
#define FOUT_IO_PORT_SEL1      P2SEL1                         /**< Portzuordnung Ausgangsregister */
#define FOUT_IO                BIT0                           /**< Pin Zuordnung  */
#define FOUT_IO_SET_SET_AS_OUT FOUT_IO_PORT_DIR |= FOUT_IO    /**< Pin als Ausgang setzen */
#define FOUT_IO_SET_SET_AS_IN  FOUT_IO_PORT_DIR &= ~(FOUT_IO) /**< Pin als Eingang setzen */
#define FOUT_IO_SET_LOW        FOUT_IO_PORT_OUT &= ~(FOUT_IO) /**< Pin Zustand auf Low setzen */
#define FOUT_IO_SET_HIGH       FOUT_IO_PORT_OUT |= FOUT_IO    /**< Pin Zustand auf High setzen */

#define FOUT_FREQ_RANGE_0   (0.0F)
#define FOUT_FREQ_RANGE_1   (1.0F)
#define FOUT_FREQ_RANGE_50  (50.0F)
#define FOUT_FREQ_RANGE_250 (250.0F)
#define FOUT_FREQ_RANGE_366 (366.0F)

#define FOUT_EN_IO_PORT_DIR   P2DIR                                /**< Portzuordnung Richtungsregister */
#define FOUT_EN_IO_PORT_OUT   P2OUT                                /**< Portzuordnung Ausgangsregister */
#define FOUT_EN_IO            BIT4                                 /**< Pin Zuordnung  */
#define FOUT_EN_IO_SET_AS_OUT FOUT_EN_IO_PORT_DIR |= FOUT_EN_IO    /**< Pin als Ausgang setzen */
#define FOUT_EN_IO_SET_LOW    FOUT_EN_IO_PORT_OUT &= ~(FOUT_EN_IO) /**< Pin Zustand auf Low setzen */
#define FOUT_EN_IO_SET_HIGH   FOUT_EN_IO_PORT_OUT |= FOUT_EN_IO    /**< Pin Zustand auf High setzen */


static void init_clk(void);


static inline void e_comp_enable(const uint8_t div) {

    FIN1_COMPARATOR_CTL_REG = CPPSEL_0;                    // Select C0 as input for V+ terminal
    FIN1_COMPARATOR_CTL_REG |= CPNSEL1 | CPNSEL2;          // Select DAC as input for V- terminal
    FIN1_COMPARATOR_CTL_REG |= CPPEN | CPNEN;              // Enable eCOMP input
    FIN1_COMPARATOR_DAC_CTL_REG |= CPDACEN;                // Select VCC as ref and enable DAC
    FIN1_COMPARATOR_DAC_DATA_REG |= div;                   // CPDACBUF1=VCC(3V3) *31/64
    FIN1_COMPARATOR_CTL1_REG |= CPEN | CPMSEL_0 | CPHSEL_1;// Turn on eCOMP, in high power mode / Hysteresis set to 10mV
}

static inline void e_comp_disable(void) {

    FIN1_COMPARATOR_CTL1_REG &= ~CPEN;
}

bool Frequencyin_set_DAC_reference_divider(const uint8_t div) {
    e_comp_disable();
    if (div > 63) return true;
    else {
        e_comp_enable(div);
        return false;
    }
}


void Frequencyoutput_init(void) {

    FOUT_IO_PORT_SEL0 |= FOUT_IO; // P2.0 CCRx Function
    FOUT_IO_PORT_SEL1 &= ~FOUT_IO;// P2.0 CCRx Function
    FOUT_IO_SET_SET_AS_OUT;
    FOUT_IO_SET_HIGH;
    FOUT_EN_IO_SET_AS_OUT;
    FOUT_EN_IO_SET_HIGH;
}

static inline uint16_t Frequencyout_get_top_register_value(const float clkSrc, CLOCK_DIV div, const float freq) {

    if (freq == 0.0F) return 0;
    const float ticks = (clkSrc / (float)div) / freq;
    return (uint16_t)(ticks - 1.0F);
}

static inline uint16_t Frequencyout_get_compare_register_value(const uint16_t ticks) {
    return (uint16_t)(ticks / 2);
}

void Frequencyoutput_set_freq(const float freq) {

    uint16_t topRegValue = 0;
    uint16_t ctlBitMask  = 0;
    uint16_t cctlBitMask = FOUT_TIMER_PWM_MODE_TOGGLE_SET;
    float    clk         = SMCLK;

    CLOCK_DIV div = CLOCK_DIV_1;

    if (freq == FOUT_FREQ_RANGE_0) {
        cctlBitMask = FOUT_TIMER_PWM_MODE_OUT;
        cctlBitMask &= ~OUT;
    } else if (freq < FOUT_FREQ_RANGE_1) {
        clk        = ACLK;
        div        = CLOCK_DIV_8;
        ctlBitMask = TBSSEL__ACLK | MC_1 | ID_3;
    } else if (freq < FOUT_FREQ_RANGE_50) {
        clk        = ACLK;
        ctlBitMask = TBSSEL__ACLK | MC_1 | ID_0;
    } else if (freq < FOUT_FREQ_RANGE_366) {
        div        = CLOCK_DIV_8;
        ctlBitMask = TBSSEL__SMCLK | MC_1 | ID_3;
    } else {
        ctlBitMask = TBSSEL__SMCLK | MC_1 | ID_0;
    }

    topRegValue         = Frequencyout_get_top_register_value(clk, div, freq);
    FOUT_TIMER_CCR0_REG = topRegValue;
    FOUT_TIMER_CCR1_REG = Frequencyout_get_compare_register_value(topRegValue);
    FOUT_TIMER_CTL_REG  = ctlBitMask;
    FOUT_TIMER_CCTL_REG = cctlBitMask;
}

static void init_clk(void) {

    WDTCTL = WDTPW | WDTHOLD;// Stop WDT
    PM5CTL0 &= ~LOCKLPM5;    // Disable the GPIO power-on default high-impedance
                             // mode to activate previously configured port settings
                             // Configure reference
    PMMCTL0_H = PMMPW_H;     // Unlock the PMM registers
    PMMCTL2 |= INTREFEN;     // Enable internal reference
    while (!(PMMCTL2 & REFGENRDY))
        ;// Poll till internal reference settles


    FRCTL0 = FRCTLPW | NWAITS_2;// FRAM waitstate for usage with 24MHz
    P2SEL1 |= BIT6 | BIT7;      // P2.6~P2.7: crystal pins
    do {
        CSCTL7 &= ~(XT1OFFG | DCOFFG);// Clear XT1 and DCO fault flag
        SFRIFG1 &= ~OFIFG;
    } while (SFRIFG1 & OFIFG);// Test oscillator fault flag

    __bis_SR_register(SCG0); // disable FLL
    CSCTL3 |= SELREF__XT1CLK;// Set XT1CLK as FLL reference source

    CSCTL1 = DCORSEL_7;                                // DCO Range = 24MHz
    CSCTL2 = FLLD_0 + BOARD_CS_MCLK_FLLREF_RATIO_24MHZ;// DCOCLKDIV = 1 32768Hz*n = SMCLK

    __delay_cycles(3);
    __bic_SR_register(SCG0);                 // enable FLL
    CSCTL4 = SELMS__DCOCLKDIV | SELA__XT1CLK;// Set ACLK = XT1CLK = 32768Hz
                                             // DCOCLK = MCLK and SMCLK source
    CSCTL5 |= DIVM_0 | DIVS_0;               // MCLK = DCOCLK
                                             // SMCLK = MCLK

    __delay_cycles(1000);// Allow clock system to settle
}


int main(void) {

    init_clk();

    FIN1_IO_PORT_SEL0 |= FIN1_IO;// Select eCOMP input function on P1.0
    FIN1_IO_PORT_SEL1 |= FIN1_IO;

    Frequencyin_set_DAC_reference_divider(32);
    Frequencyoutput_init();

    while (1) {

        Frequencyoutput_set_freq(100.0F);
    };
}
